package Client;

import org.apache.log4j.Logger;
import org.apache.log4j.FileAppender;
import org.apache.log4j.PatternLayout;

import java.io.IOException;
import java.util.*;

import Client.Util.BoardGraph;
import Client.Util.Graph;
import Shared.Constants;

/**
 * User: Naeem
 * Date: Jan 31, 2005
 * Time: 3:55:28 PM
 */
public class ABGraph extends Client{
    long timeStamp;

    /**
     * Constraint on how much the MinMax algorithm will go on
     */
    final int DEPTH = 10;

    /**
     * This field will be set to next play
     */
    int play = -1;

    /**
     * Show if play has set
     */
    boolean hasSet = false;

    /**
     * Logger using Log4j
     */
    static Logger logger = Logger.getLogger(ABGraph.class) ;
    private Random random = new Random(System.currentTimeMillis());

    ArrayList vertical = new ArrayList(18);

    public ABGraph(String fileName) throws IOException {
        super(fileName);

        for(int i = 46; i < 55; i++)
            vertical.add(new Integer(i));

        for(int i = 136; i < 145; i++)
            vertical.add(new Integer(i));

        logger.addAppender(new FileAppender(new PatternLayout("%r [%t] %-5p %c  - %m%n"), "ABGrph.log" , false));
    }

    public ABGraph(int n , int color) {
        super(n , color);
    }

    public int doPlay() {
        hasSet = false;
        timeStamp = System.currentTimeMillis();

        logger.debug("************* Next Round *************" + board);

        BoardGraph state = new BoardGraph(board);
        int size = board.getSize();
        logger.info(" Beta: " + (size - 1) * (size - 1));

//        vertical.removeAll(state.markedLines);
//
//        if(vertical.size() != 0) {
//            int vert = divider(state);
//
//            if(vert != -1)
//                return vert;
//        }

        int ehsan = lessStupidMove(state);

        if(!isFull(state))
            return ehsan;

        try {
            bigBrother(state, -1, (size - 1) * (size - 1), 0);
        } catch(TimeOutException toe) {
            logger.info(" TimeOut");
            return ehsan;
        }

        if(hasSet) {
            logger.info(" >> My play: " + play);
            return play;
        }

        if(state.count(init.color) > (state.getSize() - 1) * (state.getSize() - 1) / 2) {
            return ehsan;
        }

        try {
            maxValue(state, -1, (size - 1) * (size - 1), 0);
        } catch(TimeOutException toe) {
            logger.info(" TimeOut");
            return ehsan;
        }

        if(!hasSet)
            logger.fatal(" Not yet set!!");

        logger.info(" >> My play: " + play);
        return play;
    }

    protected int divider(BoardGraph state) {
        int evaluate = 1;
        int i;

         do {
             int index = random.nextInt(vertical.size());

            Integer tmp = ((Integer)vertical.get(index));
            i = tmp.intValue();
            vertical.remove(index);

            state.mark(i, init.color);
            evaluate = evaluate(state);
            state.undo(i);

             if(evaluate == 0)
                 break;
        }while(vertical.size() != 0);

        if(evaluate != 0)
            return -1;
        else
            return i;
    }

    protected void bigBrother(BoardGraph state, int alpha, int beta, int level) {
        int color = init.color;
        state.compute();
        Collection fixeds = new HashSet(state.getFixeds());
        Collection options = new HashSet(state.getOptions());
        Graph minop = state.getMinOption();

        if(fixeds.size() != 0) {
            for (Iterator iterator = fixeds.iterator(); iterator.hasNext();) {
                Graph f = (Graph) iterator.next();
                play = f.getUnSafeLine();
                hasSet = true;
                return;
            }
        }

        if(options.size() != 0) {
            for (Iterator iterator = options.iterator(); iterator.hasNext();) {
                Graph o = (Graph) iterator.next();
                play = o.getUnSafeLine();
                hasSet = true;
                return;
            }
        }

        if(minop != null) {
            if(minop.getSafeLineNo() != -1) {
                play = minop.getSafeLineNo();
                hasSet = true;
                return;
            }

            state.markGraph(minop, color, false);
            int all = maxValue(state, alpha, beta, level);
            state.unMarkGrpah(minop);

            state.markGraph(minop, color, true);
            int turn = maxFreak(state, alpha, beta, level);
            state.unMarkGrpah(minop);

            if(all > turn)
                play = minop.getUnSafeLine();
            else
                play = minop.getTrunLine();

            hasSet = true;
            return;
        }
    }

    protected int maxManager(BoardGraph state, int alpha, int beta, int level) {
        if(System.currentTimeMillis() - timeStamp > 200)
            throw new TimeOutException();

        int color = init.color;
        state.compute();
        Collection fixeds = new HashSet(state.getFixeds());
        Collection options = new HashSet(state.getOptions());
        Graph minop = state.getMinOption();

        if(fixeds.size() != 0) {
            for (Iterator iterator = fixeds.iterator(); iterator.hasNext();) {
                Graph f = (Graph) iterator.next();
                state.markGraph( f, color, false);
            }
        }

        if(options.size() != 0) {
            for (Iterator iterator = options.iterator(); iterator.hasNext();) {
                Graph o = (Graph) iterator.next();
                state.markGraph(o, color, false);
            }
        }

        int result;

        if(minop != null) {
            state.markGraph(minop, color, false);
            int all = maxValue(state, alpha, beta, level);
            state.unMarkGrpah(minop);

            state.markGraph(minop, color, true);
            int turn = maxFreak(state, alpha, beta, level);
            state.unMarkGrpah(minop);

            result = Math.max(all, turn);
        } else {
            result = maxValue(state, alpha, beta, level);
        }

        if(options.size() != 0) {
            for (Iterator iterator = options.iterator(); iterator.hasNext();) {
                Graph o = (Graph) iterator.next();
                state.unMarkGrpah(o);
            }
        }

        if(fixeds.size() != 0) {
            for (Iterator iterator = fixeds.iterator(); iterator.hasNext();) {
                Graph f = (Graph) iterator.next();
                state.unMarkGrpah(f);
            }
        }

        return result;
    }

    protected int maxFreak(BoardGraph state, int alpha, int beta, int level) {
        int min = minManager(state, alpha, beta, level);

        if( alpha < min)
            alpha = min;

        if( beta <= alpha)
            return beta;

        return alpha;
    }

    protected int maxValue(BoardGraph state, int alpha, int beta, int level) {
        if ( cutoffTest(state, level))
            return eval(state);

        int color = init.color;
        Set candidates = state.getFreeLines();
        int min;

        for (Iterator it = candidates.iterator(); it.hasNext();) {
            Integer linenoInt = (Integer) it.next();
            int lineno = linenoInt.intValue();

            if (!state.markedLines.contains(linenoInt)) {
                int me = state.count(color);

                state.mark(lineno, color);

                if(me < state.count(color))
                    min = minFreak(state, alpha, beta, level);
                else
                    min = minManager(state, alpha, beta, level + 1);

                if(level == 0)
                    logger.info(" Line: " + lineno + " Value: " + min + " Alpha: " + alpha + " Beta: " + beta);

                if( alpha < min) {
                    alpha = min;

                    if( level == 0) {
                        play = lineno;
                        hasSet = true;
                    }
                }

                state.undo(lineno);

                if( beta <= alpha)
                    return beta;
            }
        }

        return alpha;
    }

    protected int minManager(BoardGraph state, int alpha, int beta, int level) {
        int color;

        if(init.color == Constants.BLUE)
            color = Constants.RED;
        else
            color = Constants.BLUE;

        state.compute();
        Collection fixeds = new HashSet(state.getFixeds());
        Collection options = new HashSet(state.getOptions());
        Graph minop = state.getMinOption();

        if(fixeds.size() != 0) {
            for (Iterator iterator = fixeds.iterator(); iterator.hasNext();) {
                Graph f = (Graph) iterator.next();
                state.markGraph( f, color, false);
            }
        }

        if(options.size() != 0) {
            for (Iterator iterator = options.iterator(); iterator.hasNext();) {
                Graph o = (Graph) iterator.next();
                state.markGraph(o, color, false);
            }
        }

        int result;

        if(minop != null) {
            state.markGraph(minop, color, false);
            int all = minValue(state, alpha, beta, level);
            state.unMarkGrpah(minop);

            state.markGraph(minop, color, true);
            int turn = minFreak(state, alpha, beta, level);
            state.unMarkGrpah(minop);

            result = Math.min(all, turn);
        } else {
            result = minValue(state, alpha, beta, level);
        }

        if(options.size() != 0) {
            for (Iterator iterator = options.iterator(); iterator.hasNext();) {
                Graph o = (Graph) iterator.next();
                state.unMarkGrpah(o);
            }

        }

        if(fixeds.size() != 0) {
            for (Iterator iterator = fixeds.iterator(); iterator.hasNext();) {
                Graph f = (Graph) iterator.next();
                state.unMarkGrpah(f);
            }
        }

        return result;
    }

    protected int minFreak(BoardGraph state, int alpha, int beta, int level) {
        int max = maxManager(state, alpha, beta, level);

        if( beta > max)
            beta = max;

        if (beta <= alpha)
            return alpha;

        return beta;
    }

    protected int minValue(BoardGraph state, int alpha, int beta, int level) {
        if ( cutoffTest(state, level))
            return eval(state);

        int color;
        Set candidates = state.getFreeLines();
        int max;

        if(init.color == Constants.BLUE)
            color = Constants.RED;
        else
            color = Constants.BLUE;

        for (Iterator it = candidates.iterator(); it.hasNext();) {
            Integer linenoInt = (Integer) it.next();
            int lineno = linenoInt.intValue();

            if (!state.markedLines.contains(linenoInt)) {
                int oponent = state.count(color);

                state.mark(lineno, color);

                if(oponent < state.count(color))
                    max = maxFreak(state, alpha, beta, level);
                else
                    max = maxManager(state, alpha, beta, level + 1);

                if( beta > max)
                    beta = max;

                state.undo(lineno);

                if (beta <= alpha)
                    return alpha;
            }
        }

        return beta;
    }

    //Todo revise it
    protected int eval(BoardGraph state) {
        return state.count( init.color);
    }

    protected boolean cutoffTest(BoardGraph state, int level) {
        int markedSize = state.markedLines.size();
        int size = state.getSize();
        int cellsize = (size - 1) * (size - 1) / 2;
        size = (size * (size - 1)) * 2;

        if( state.getFreeLines().size() == 0 || markedSize == size || level > DEPTH) // || state.count(init.color) > cellsize
            return true;

        return false;
    }
    public int myEval(BoardGraph b) {
        //System.out.println("My Eval ");
        int index = 0 ;
        int max = 0 ;
        for (int i = 1 ; i < b.getMaxLine()+1 ; i++) {
            if (!b.markedLines.contains(new Integer(i))) {
                int tmp = b.checkLine(i) ;
                if (tmp  > 0 ) {
//                    System.out.println("Here!");
                    //System.out.println("I is " + i );
                    b.mark(i , Constants.BLUE) ;
                    index = myEval(b) + tmp;
                    b.undo(i);
                    return index ;

                }
            }
        }

        return 0 ;

    }
    public int getStupidMove(BoardGraph board) {
        int index = 0 ;
        int max = 0 ;
        for (int i = 1 ; i < board.getMaxLine()+1 ; i++) {
            if (!board.markedLines.contains(new Integer(i))) {
                int tmp = board.checkLine(i) ;
                if (tmp  > max ) {
                    //System.out.println("I is " + i );
                    index = i ;
                    max = tmp ;

                }
            }
        }
        if (index > 0 ) {
//            System.out.println(" Ready cell ");
            return index ;
        }

        int min = 10000 ;
        boolean withOption = true;
        ArrayList candid = new  ArrayList();
        for (int i = board.getMaxLine() ; i > 0 ; i--) {

            Integer integer = new Integer(i) ;
            if (board.markedLines.contains(integer))
                continue;

            board.mark(integer.intValue() , Constants.BLUE) ;//Blue is not important
            int tmp =  evaluate(board) ;

            if (min > tmp || (min==tmp && (withOption && !board.hasOption))) {
                candid.clear() ;
                min = tmp ;
                candid.add(integer) ;
                withOption = board.hasOption ;
            } else if  (min == tmp && board.hasOption == withOption) {
                candid.add(integer) ;
            }

            board.undo(integer.intValue());
        }
        index = ((Integer)candid.get(random.nextInt(candid.size()))).intValue() ;
//        System.out.println("Min is " + min + " Index: " + index);
        return index ;

    }

    protected int lessStupidMove(BoardGraph state) {
        int color = init.color;
        state.compute();
        Collection fixeds = new HashSet(state.getFixeds());
        Collection options = new HashSet(state.getOptions());
        Collection generalOptions = new HashSet(state.getGeneralOptions());

        Graph minop = state.getMinOption();

        if(fixeds.size() != 0 || options.size() != 0)
            return getStupidMove(state);

        if(minop != null) {
            if(minop.getSafeLineNo() != -1)
                return minop.getSafeLineNo();

            if(state.getGraphs().size() == 1)
                return getStupidMove(state);

            state.markGraph(minop, Constants.BLUE, false);
            int evaluate = evaluate(state);
            int stupid = getStupidMove(state);
            state.unMarkGrpah(minop);

            if(evaluate > 0)
                return stupid;
            else {
                //Todo Even or Odd
                state.markGraph(minop, color, false);
                int move = getStupidMove(state);
                state.mark(move, color);
                int all = evaluate(state);
                state.undo(move);
                state.unMarkGrpah(minop);

                int turn = minop.getBestScore() - minop.getTurnScore();

                if(all < turn)
                    return minop.getUnSafeLine();
                else
                    return minop.getTrunLine();
            }
        } else if(generalOptions.size() != 0) {
            Graph go = (Graph) generalOptions.iterator().next();

            if(generalOptions.size() > 1)
                return go.getUnSafeLine();

            if(go.getSafeLineNo() != -1)
                return go.getSafeLineNo();

            BoardGraph myState = new BoardGraph(state);
            BoardGraph yourState = new BoardGraph(state);
            myState.compute();
            Graph mingo = (Graph) myState.getGeneralOptions().iterator().next();

            int unsafe = mingo.getUnSafeLine();
            int changeTurn = mingo.getTrunLine();

            yourState.mark(changeTurn, color);
            int turn = evaluate(yourState);
            yourState.undo(changeTurn) ;

            myState.markMagic(mingo, color);
            int move = getStupidMove(myState);
            myState.mark(move, color);
            int all = evaluate(myState);
            myState.undo(move) ;

            if(all < turn)
                return unsafe;
            else
                return changeTurn;
        }

        return getStupidMove(state);
    }

    public int evaluate(BoardGraph board) {
        int max = 0 ;

        board.compute() ;
        max = myEval(board) ; ;
//        if (max > 0)
//            System.out.println("max " + max);
        return max ;
/*

        int fCells = 0 ;
        int oCells  = 0 ;

        for (Iterator iterator = board.getGraphs().iterator(); iterator.hasNext();) {
            Graph graph = (Graph) iterator.next();
            if(graph.getType() == Graph.GENERALTYPE)
                max += graph.totalRTO ;
            else if(graph.getType() == Graph.OPTIONTYPE)
                oCells += graph.getBestScore() ;
            else if(graph.getType() == Graph.FIXEDTYPE)
                fCells += graph.getBestScore() ;
        }

        if (oCells == 0)
            return fCells + max;
        else
            return fCells + oCells + max  ;
        //TODO make option types more prior
*/

    }


    public boolean isFull(BoardGraph boardGraph) {
        int semiUsed = 0, fullUsed = 0;
        for (Iterator iterator = boardGraph.markedLines.iterator(); iterator.hasNext();) {
            Integer integer = (Integer) iterator.next();
            int ln  = boardGraph.checkLine(integer.intValue()) ;
            if (ln == 2)
                fullUsed++ ;
            else if (ln == 1) {
                if (boardGraph.isBorder(integer.intValue()))
                    fullUsed++ ;
                else
                    semiUsed++ ;
            }
        }

        int remain =boardGraph.markedLines.size()  - fullUsed - semiUsed / 2;
        int all = boardGraph.getMaxLine() - boardGraph.markedLines.size() ;

        float ratio = ((float)remain) / all ;
        logger.info("Ratio " + ratio + " marked " + boardGraph.markedLines.size() + " full " + fullUsed +
                " semi "+ semiUsed);
        logger.info("Remain " + remain + " all " + all) ;
        return ratio > 0.7 ;
    }

}
