package Shared;

import Shared.Messages.SimpleParse;

import java.util.*;

import org.apache.log4j.Logger;
import Client.ABGraph;

/**
 * Created by IntelliJ IDEA.
 * User: sehsanm
 * Date: Dec 13, 2004
 * Time: 10:45:03 PM
 * To change this template use Options | File Templates.
 */
public class Board {
    static Logger ablogger = Logger.getLogger(ABGraph.class);
    /**
     * N is number of points in each direction.
     */
    protected int n;

    /**
     * Numbers of lines which has been assigned. Line numbers start from 1.
     */
    public Set markedLines = new HashSet();

    /**
     * Cells of board. zero indexed.
     */
    public int[][] cells;

    /**
     *
     * @param size Number of points in each side.
     */
    public Board(int size) {
        n = size;
        cells = new int[n - 1][n - 1];
        for (int i = 0; i < n - 1; i++)
            for (int j = 0; j < n - 1; j++)
                cells[i][j] = Constants.FREE;
    }

    public Board(Board other) {
        n = other.n;

        for (Iterator iterator = other.markedLines.iterator(); iterator.hasNext();) {
            Integer integer = (Integer) iterator.next();

            markedLines.add(integer);
        }

        cells = new int[n - 1][n - 1];

        for (int i = 0; i < n - 1; i++) {
            for (int j = 0; j < n - 1; j++) {
                cells[i][j] = other.cells[i][j];
            }
        }
    }

    /**
     *
     * @return size of the board(Number of points in each side).
     */
    public int getSize() {
        return n;
    }

    /**
     * Add a marked line. If the Line exist does not any action.
     * @param lineno Line number to be marked.
     * @param turn Turn of the player (RED or BLUE)
     * @return -1 if illegal action is tacken thoerwise it returns number
     * of new cells occupied.
     */
    public int mark(int lineno, int turn) {

        Integer tmp = new Integer(lineno);
        if (markedLines.contains(tmp) || lineno > (n * (n - 1) * 2) || lineno < 1) {
            ablogger.debug("Rejecting " + lineno + " n: " + n + " Contain: " + markedLines.contains(tmp));
            throw new NullPointerException("Reject " + lineno);
//            return -1;
        }
        markedLines.add(tmp);
        return updateLine(lineno, turn);
    }

    /**
     *
     * @return True if the game is finished.
     */
    public boolean isFinished() {
        return (markedLines.size() == 2 * (n - 1) * n);
    }

    /**
     * Checks if a line is vertical.
     * @param ln Line number
     * @return True if the line is vertical. otherwise 0.
     */
    protected boolean isVertical(int ln) {
        return (ln <= (n - 1) * n);
    }

    /**
     *Ln should be number of a vertical line.
     * @param ln  Line number should be vertical.
     * @return Number of point row which this horizantal line's upper point is on that(zero indexed).
     */
    protected int rowVert(int ln) {
        return (ln - 1) % (n - 1);
    }

    /**
     * @param ln Line number. Should be a vertical line number.
     * @return Number of point columns which this line on that. (zero indexed).
     */
    protected int colVert(int ln) {
        return (ln - 1) / (n - 1);
    }

    /**
     *
     * @param ln Line number. should be horizantal.
     * @return Number of row which this line is on that(zero indexed).
     */
    protected int rowHor(int ln) {
        ln -= n * (n - 1);
        return ((ln - 1) / (n - 1));
    }

    /**
     *
     * @param ln  Line number(should be horizantal).
     * @return Number of column which left part of the line is on that.
     */
    protected int colHor(int ln) {
        ln -= n * (n - 1);
        return ((ln - 1) % (n - 1));
    }

    /**
     * Checks if the line is added does it produce any cell or not ?
     * @param ln Line number
     * @return   Number of rectangles that is produced if the line is added (0 , 1 or 2).Note
     * that this function does not check if the line is marked or not. It's duty of programmer.
     */
    public int checkLine(int ln) {
        if (isVertical(ln))
            return checkVertical(ln);
        else
            return checkHorizantal(ln);

    }

    /**
     * Checks if the hirizantal line is added does it produce any cell or not ?
     * @param ln Line number
     * @return   Number of rectangles that is produced if the line is added (0 , 1 or 2).Note
     * that this function does not check if the line is marked or not. It's duty of programmer.
     */
    protected int checkHorizantal(int ln) {
        int up = ln - (n - 1);
        if (isVertical(up))
            up = -1;
        int down = ln + (n - 1);
        int upLeft = colHor(ln) * (n - 1) + rowHor(ln);
        int upRight = (colHor(ln) + 1) * (n - 1) + rowHor(ln);
        int downLeft = upLeft + 1;
        int downRight = upRight + 1;


        HashSet upRect = new HashSet(4);
        upRect.add(new Integer(up));
        upRect.add(new Integer(upLeft));
        upRect.add(new Integer(upRight));

        HashSet downRect = new HashSet(4);
        downRect.add(new Integer(down));
        downRect.add(new Integer(downLeft));
        downRect.add(new Integer(downRight));

        int ret = 0;
        if (markedLines.containsAll(downRect))
            ret++;
        if (markedLines.containsAll(upRect))
            ret++;
        return ret;
    }

    /**
     * Checks if the vertical line is added does it produce any cell or not ?
     * @param ln Line number
     * @return   Number of rectangles that is produced if the vertical line is added (0 , 1 or 2).Note
     * that this function does not check if the line is marked or not. It's duty of programmer.
     */
    protected int checkVertical(int ln) {
        int leftLine = ln - (n - 1);
        int rightLine = ln + (n - 1);
        if (!isVertical(rightLine))
            rightLine = -1;
        int upLeft = (n * (n - 1) + rowVert(ln) * (n - 1) + colVert(ln));
        int downLeft = (n * (n - 1) + (rowVert(ln) + 1) * (n - 1) + colVert(ln));
        int upRight = upLeft + 1;
        int downRight = downLeft + 1;

        HashSet leftRect = new HashSet(4);
        leftRect.add(new Integer(leftLine));
        leftRect.add(new Integer(upLeft));
        leftRect.add(new Integer(downLeft));

        HashSet rightRect = new HashSet(4);
        rightRect.add(new Integer(rightLine));
        rightRect.add(new Integer(upRight));
        rightRect.add(new Integer(downRight));

        int ret = 0;
        if (markedLines.containsAll(leftRect))
            ret++;
        if (markedLines.containsAll(rightRect))
            ret++;
        return ret;
    }

    /**
     *Checks if the specified line has made a new rectangle. If it has
     * it adds rectangles to players occupied cells.
     * @param lineno Line number.
     * @param turn Player who has marked the line.
     * @return Number of rectangles which line has made.
     */
    protected int updateLine(int lineno, int turn) {
        if (isVertical(lineno))
            return updateVertical(lineno, turn);
        else
            return updateHorizantal(lineno, turn);
    }

    /**
     *Checks if the specified Horizantal line has made a new rectangle. If it has
     * it adds rectangles to players occupied cells.
     * @param lineno Line number.
     * @param turn Player who has marked the line.
     * @return Number of rectangles which line has made.
     */
    protected int updateHorizantal(int lineno, int turn) {

        int up = lineno - (n - 1);
        if (isVertical(up))
            up = -1;
        int down = lineno + (n - 1);
        int upLeft = colHor(lineno) * (n - 1) + rowHor(lineno);
        int upRight = (colHor(lineno) + 1) * (n - 1) + rowHor(lineno);
        int downLeft = upLeft + 1;
        int downRight = upRight + 1;


        HashSet upRect = new HashSet(4);
        upRect.add(new Integer(lineno));
        upRect.add(new Integer(up));
        upRect.add(new Integer(upLeft));
        upRect.add(new Integer(upRight));

        HashSet downRect = new HashSet(4);
        downRect.add(new Integer(lineno));
        downRect.add(new Integer(down));
        downRect.add(new Integer(downLeft));
        downRect.add(new Integer(downRight));

        int ret = 0;

        if (markedLines.containsAll(upRect)) {
            ret++;
            cells[rowHor(lineno) - 1][colHor(lineno)] = turn;
        }
        if (markedLines.containsAll(downRect)) {
            ret++;
            cells[rowHor(lineno)][colHor(lineno)] = turn;

        }

        return ret;


    }

    /**
     *Checks if the specified vertical line has made a new rectangle. If it has
     * it adds rectangles to players occupied cells.
     * @param ln Line number.
     * @param turn Player who has marked the line.
     * @return Number of rectangles which line has made.
     */
    protected int updateVertical(int ln, int turn) {
        //Left Rect
        int leftLine = ln - (n - 1);
        int rightLine = ln + (n - 1);
        if (!isVertical(rightLine))
            rightLine = -1;
        int upLeft = (n * (n - 1) + rowVert(ln) * (n - 1) + colVert(ln));
        int downLeft = (n * (n - 1) + (rowVert(ln) + 1) * (n - 1) + colVert(ln));
        int upRight = upLeft + 1;
        int downRight = downLeft + 1;

        HashSet leftRect = new HashSet(4);
        leftRect.add(new Integer(leftLine));
        leftRect.add(new Integer(ln));
        leftRect.add(new Integer(upLeft));
        leftRect.add(new Integer(downLeft));

        HashSet rightRect = new HashSet(4);
        rightRect.add(new Integer(ln));
        rightRect.add(new Integer(rightLine));
        rightRect.add(new Integer(upRight));
        rightRect.add(new Integer(downRight));

        int ret = 0;
        if (markedLines.containsAll(leftRect)) {
            ret++;
            cells[rowVert(ln)][colVert(ln) - 1] = turn;
        }
        if (markedLines.containsAll(rightRect)) {
            ret++;
            cells[rowVert(ln)][colVert(ln)] = turn;

        }
        return ret;
    }

    /**
     * Try to undo a maked line an its rectangles
     * @param lineno Line number
     */
    public void undo(int lineno) {
        Integer tmp = new Integer(lineno);

        if (!markedLines.contains(tmp) || lineno > (n * (n - 1) * 2) || lineno < 1) {

            ablogger.debug("Rejecting " + lineno + " State: " + this);
            throw new NullPointerException("Reject");
//            return;
        }

        undoLine(tmp.intValue());
        markedLines.remove(tmp);
    }

    /**
     * Checks if the specified line has removed (a) rectangle(s). If it has
     * it frees rectangles.
     * @param lineno Line number.
     */
    protected void undoLine(int lineno) {
        if (isVertical(lineno))
            undoVertical(lineno);
        else
            undoHorizantal(lineno);
    }

    /**
     * Checks if the specified horizontal line has removed (a) rectangle(s). If it has
     * it frees rectangles.
     * @param lineno Line number.
     */
    protected void undoHorizantal(int lineno) {

        int up = lineno - (n - 1);
        if (isVertical(up))
            up = -1;
        int down = lineno + (n - 1);
        int upLeft = colHor(lineno) * (n - 1) + rowHor(lineno);
        int upRight = (colHor(lineno) + 1) * (n - 1) + rowHor(lineno);
        int downLeft = upLeft + 1;
        int downRight = upRight + 1;

        HashSet upRect = new HashSet(4);
        upRect.add(new Integer(lineno));
        upRect.add(new Integer(up));
        upRect.add(new Integer(upLeft));
        upRect.add(new Integer(upRight));

        HashSet downRect = new HashSet(4);
        downRect.add(new Integer(lineno));
        downRect.add(new Integer(down));
        downRect.add(new Integer(downLeft));
        downRect.add(new Integer(downRight));

        if (markedLines.containsAll(upRect))
            cells[rowHor(lineno) - 1][colHor(lineno)] = Constants.FREE;

        if (markedLines.containsAll(downRect))
            cells[rowHor(lineno)][colHor(lineno)] = Constants.FREE;
    }

    /**
     * Checks if the specified verticla line has removed (a) rectangle(s). If it has
     * it frees rectangles.
     * @param lineno Line number.
     */
    protected void undoVertical(int lineno) {
        //Left Rect
        int leftLine = lineno - (n - 1);
        int rightLine = lineno + (n - 1);

        if (!isVertical(rightLine))
            rightLine = -1;
        int upLeft = (n * (n - 1) + rowVert(lineno) * (n - 1) + colVert(lineno));
        int downLeft = (n * (n - 1) + (rowVert(lineno) + 1) * (n - 1) + colVert(lineno));
        int upRight = upLeft + 1;
        int downRight = downLeft + 1;

        HashSet leftRect = new HashSet(4);
        leftRect.add(new Integer(leftLine));
        leftRect.add(new Integer(lineno));
        leftRect.add(new Integer(upLeft));
        leftRect.add(new Integer(downLeft));

        HashSet rightRect = new HashSet(4);
        rightRect.add(new Integer(lineno));
        rightRect.add(new Integer(rightLine));
        rightRect.add(new Integer(upRight));
        rightRect.add(new Integer(downRight));

        if (markedLines.containsAll(leftRect))
            cells[rowVert(lineno)][colVert(lineno) - 1] = Constants.FREE;

        if (markedLines.containsAll(rightRect))
            cells[rowVert(lineno)][colVert(lineno)] = Constants.FREE;
    }

    /**
     *
     * @param color The color of player RED or BLUE
     * @return number of occupied cells  by the player
     */
    public int count(int color) {
        int count = 0;

        for (int i = 0; i < n - 1; i++) {
            for (int j = 0; j < n - 1; j++) {
                if (cells[i][j] == color) {
                    count++;
                }
            }
        }

        return count;
    }


    /**
     * Make the string corresponding current state of the board. This string is used
     * to transfer current state of the board to players and viewer.Format of the
     * tranfered data is  (Marked lines sequence)(Red marked cells)(Blue marked cells).
     * For example (17 , 8 , 20 ,5) (4) ()
     * @return String representing the state.
     */
    public String toString() {
        String ret = "(";
        boolean first = true;
        for (Iterator iterator = markedLines.iterator(); iterator.hasNext();) {
            Integer integer = (Integer) iterator.next();
            if (!first)
                ret += " ," + integer;
            else
                ret += integer;
            first = false;
        }
        ret += ")(";

        first = true;
        for (int i = 0; i < n - 1; i++)
            for (int j = 0; j < n - 1; j++)
                if (cells[i][j] == Constants.RED) {
                    int integer = i * (n - 1) + j;
                    if (!first)
                        ret += " ," + integer;
                    else
                        ret += integer;
                    first = false;

                }
        ret += ")(";
        first = true;
        for (int i = 0; i < n - 1; i++)
            for (int j = 0; j < n - 1; j++)
                if (cells[i][j] == Constants.BLUE) {
                    int integer = i * (n - 1) + j;
                    if (!first)
                        ret += " ," + integer;
                    else
                        ret += integer;
                    first = false;

                }

        ret += ")";


        return ret;
    }

    /**
     * ReConstructs the board from the string produced by toString().
     * @param data String reperesenting the board.
     */
    public void parseString(String data) {
        markedLines.clear();
        markedLines.addAll(SimpleParse.getFirstVector(data));
        data = SimpleParse.deleteFirstVector(data);
        ArrayList r = SimpleParse.getFirstVector(data);
        for (int i = 0; i < n - 1; i++)
            for (int j = 0; j < n - 1; j++)
                cells[i][j] = Constants.FREE;
        for (Iterator iterator = r.iterator(); iterator.hasNext();) {
            Integer integer = (Integer) iterator.next();
            int x = integer.intValue();
            cells[x / (n - 1)][x % (n - 1)] = Constants.RED;
        }

        data = SimpleParse.deleteFirstVector(data);
        r = SimpleParse.getFirstVector(data);
        for (Iterator iterator = r.iterator(); iterator.hasNext();) {
            Integer integer = (Integer) iterator.next();
            int x = integer.intValue();
            cells[x / (n - 1)][x % (n - 1)] = Constants.BLUE;
        }


    }

    /**
     * return the lineXY representation of the lineNo.
     * @param lineNo  Line number.
     * @return Cordinates of line head and tail.
     */
    public LineXY getLine(int lineNo) {
        if (lineNo > (n * (n - 1) * 2))
            return null;
        LineXY ret = new LineXY();
        if (isVertical(lineNo)) {
            ret.x1 = colVert(lineNo);
            ret.y1 = rowVert(lineNo);
            ret.x2 = ret.x1;
            ret.y2 = ret.y1 + 1;
        } else {
            ret.x1 = colHor(lineNo);
            ret.y1 = rowHor(lineNo);
            ret.x2 = ret.x1 + 1;
            ret.y2 = ret.y1;
        }
        return ret;

    }

    /**
     *
     * @return Set of lineXY cordinates of markedLines.
     */
    public Set getLines() {
        HashSet ret = new HashSet();
        for (Iterator iterator = markedLines.iterator(); iterator.hasNext();) {
            int i = ((Integer) iterator.next()).intValue();
            ret.add(getLine(i));

        }
        return ret;
    }

    /**
     *
     * @return Maximum number of lines in the board.
     */
    public int getMaxLine() {
        return 2 * n * (n - 1);
    }

    public boolean isBorder(int ln) {
        if (isVertical(ln)) {
            if (colVert(ln) == 0 || colVert(ln) == n - 1 )
                return true ;
        } else {
            if (rowHor(ln) == 0 || rowHor(ln) == n - 1)
                return true ;
        }

        return false ;

    }


}
